__all__ = [
    'RunExposure',
    'RunFmTest',
]

import tempfile
import os
import csv
import shutil

from itertools import chain
import pandas as pd

from ..base import ComputationStep
from ..generate.keys import GenerateKeysDeterministic
from ..generate.files import GenerateFiles
from ..generate.losses import GenerateLossesDeterministic

from ...preparation.il_inputs import get_oed_hierarchy
from ...utils.exceptions import OasisException

from ...utils.data import (
    get_dataframe,
    print_dataframe, get_exposure_data,
)
from ...utils.inputs import str2bool
from ...utils.coverages import SUPPORTED_COVERAGE_TYPES
from ...utils.defaults import (
    KTOOLS_ALLOC_IL_DEFAULT,
    KTOOLS_ALLOC_RI_DEFAULT,
    KTOOLS_ALLOC_FM_MAX,
    OASIS_FILES_PREFIXES,
)


class RunExposure(ComputationStep):
    """
    Generates insured losses from preexisting Oasis files with specified
    loss factors (loss % of TIV).
    """
    step_params = [
        {'name': 'src_dir', 'flag': '-s', 'is_path': True, 'pre_exist': True, 'help': ''},
        {'name': 'run_dir', 'flag': '-r', 'is_path': True, 'pre_exist': False, 'help': ''},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'output_file', 'flag': '-f', 'is_path': True, 'pre_exist': False, 'help': '', 'type': str},
        {'name': 'loss_factor', 'flag': '-l', 'type': float, 'nargs': '+', 'help': '', 'default': [1.0]},
        {'name': 'currency_conversion_json', 'is_path': True, 'pre_exist': True, 'help': 'settings to perform currency conversion of oed files'},
        {'name': 'reporting_currency', 'help': 'currency to use in the results reported'},
        {'name': 'ktools_alloc_rule_il', 'flag': '-a', 'default': KTOOLS_ALLOC_IL_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in direct insured loss'},
        {'name': 'ktools_alloc_rule_ri', 'flag': '-A', 'default': KTOOLS_ALLOC_RI_DEFAULT, 'type': int,
         'help': 'Set the fmcalc allocation rule used in reinsurance'},
        {'name': 'output_level', 'flag': '-o', 'help': 'Keys files output format', 'choices': ['item', 'loc', 'pol', 'acc', 'port'],
         'default': 'item'},
        {'name': 'num_subperils', 'flag': '-p', 'default': 1, 'type': int,
         'help': 'Set the number of subperils returned by deterministic key generator'},
        {'name': 'coverage_types', 'type': int, 'nargs': '+', 'default': list(v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()),
         'help': 'Select List of supported coverage_types [1, .. ,4]'},
        {'name': 'fmpy', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use fmcalc python version instead of c++ version'},
        {'name': 'fmpy_low_memory', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use memory map instead of RAM to store loss array (may decrease performance but reduce RAM usage drastically)'},
        {'name': 'fmpy_sort_output', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'order fmpy output by item_id'},
        {'name': 'stream_type', 'flag': '-t', 'default': 2, 'type': int,
         'help': 'Set the IL input stream type, 2 = default loss stream, 1 = deprecated cov/item stream'},
        {"name": "hashed_group_id", "default": True, "type": str2bool, "const": False, 'nargs': '?', "help": "Hashes the group_id in the items.bin"},
        {'name': 'net_ri', 'default': True},
        {'name': 'include_loss_factor', 'default': True},
        {'name': 'print_summary', 'default': True},
    ]

    def _check_alloc_rules(self):
        alloc_ranges = {
            'ktools_alloc_rule_il': KTOOLS_ALLOC_FM_MAX,
            'ktools_alloc_rule_ri': KTOOLS_ALLOC_FM_MAX}
        for rule in alloc_ranges:
            alloc_val = getattr(self, rule)
            if (alloc_val < 0) or (alloc_val > alloc_ranges[rule]):
                raise OasisException(f'Error: {rule}={alloc_val} - Not withing valid range [0..{alloc_ranges[rule]}]')

    def run(self):
        tmp_dir = None
        src_dir = self.src_dir if self.src_dir else os.getcwd()

        if self.run_dir:
            run_dir = self.run_dir
        else:
            tmp_dir = tempfile.TemporaryDirectory()
            run_dir = tmp_dir.name

        include_loss_factor = not (len(self.loss_factor) == 1)

        self._check_alloc_rules()

        self.oasis_files_dir = src_dir
        exposure_data = get_exposure_data(self, add_internal_col=True)
        if not exposure_data.location or not exposure_data.account:
            raise OasisException(
                f'No location/exposure found in source directory "{src_dir}" - '
                'a file named `location.*` is expected'
            )
        il = bool(exposure_data.account)
        ril = all([exposure_data.ri_info, exposure_data.ri_scope, il])

        self.logger.debug('\nRunning deterministic losses (GUL=True, IL={}, RIL={})\n'.format(il, ril))

        if not os.path.exists(run_dir):
            os.makedirs(run_dir)

        # 1. Create Deterministic keys file
        keys_fp = os.path.join(run_dir, 'keys.csv')
        GenerateKeysDeterministic(
            keys_data_csv=keys_fp,
            num_subperils=self.num_subperils,
            supported_oed_coverage_types=self.coverage_types,
            exposure_data=exposure_data,
        ).run()

        # 2. Start Oasis files generation
        GenerateFiles(
            oasis_files_dir=run_dir,
            exposure_data=exposure_data,
            keys_data_csv=keys_fp,
            hashed_group_id=self.hashed_group_id,
        ).run()

        # 3. Run Deterministic Losses
        losses = GenerateLossesDeterministic(
            exposure_data=exposure_data,
            oasis_files_dir=run_dir,
            output_dir=os.path.join(run_dir, 'output'),
            include_loss_factor=include_loss_factor,
            loss_factor=self.loss_factor,
            num_subperils=self.num_subperils,
            net_ri=self.net_ri,
            ktools_alloc_rule_il=self.ktools_alloc_rule_il,
            ktools_alloc_rule_ri=self.ktools_alloc_rule_ri,
            fmpy=self.fmpy,
            fmpy_low_memory=self.fmpy_low_memory,
            fmpy_sort_output=self.fmpy_sort_output,
            il_stream_type=self.stream_type,
        ).run()

        guls_df = losses['gul']
        ils_df = losses['il']
        rils_df = losses['ri']

        # Read in the summary map
        summaries_df = get_dataframe(src_fp=os.path.join(run_dir, 'fm_summary_map.csv'), lowercase_cols=False)

        guls_df.to_csv(path_or_buf=os.path.join(run_dir, 'guls.csv'), index=False, encoding='utf-8')
        guls_df.rename(columns={'loss': 'loss_gul'}, inplace=True)

        guls_df = guls_df.merge(
            right=summaries_df,
            left_on=["item_id"],
            right_on=["agg_id"]
        )

        if include_loss_factor:
            join_cols = ["event_id", "output_id", "loss_factor_idx"]
        else:
            join_cols = ["event_id", "output_id"]
        if il:
            ils_df.to_csv(path_or_buf=os.path.join(run_dir, 'ils.csv'), index=False, encoding='utf-8')
            ils_df.rename(columns={'loss': 'loss_il'}, inplace=True)
            all_losses_df = guls_df.merge(
                how='left',
                right=ils_df,
                on=join_cols,
                suffixes=["_gul", "_il"]
            )

        if ril:
            rils_df.to_csv(path_or_buf=os.path.join(run_dir, 'rils.csv'), index=False, encoding='utf-8')
            rils_df.rename(columns={'loss': 'loss_ri'}, inplace=True)
            all_losses_df = all_losses_df.merge(
                how='left',
                right=rils_df,
                on=join_cols
            )

        oed_hierarchy = get_oed_hierarchy()
        portfolio_num = oed_hierarchy['portnum']['ProfileElementName']
        acc_num = oed_hierarchy['accnum']['ProfileElementName']
        loc_num = oed_hierarchy['locnum']['ProfileElementName']
        policy_num = oed_hierarchy['polnum']['ProfileElementName']

        if self.output_level == 'port':
            summary_cols = [portfolio_num]
        elif self.output_level == 'acc':
            summary_cols = [portfolio_num, acc_num]
        elif self.output_level == 'pol':
            summary_cols = [portfolio_num, acc_num, policy_num]
        elif self.output_level == 'loc':
            summary_cols = [portfolio_num, acc_num, loc_num]
        elif self.output_level == 'item':
            summary_cols = [
                'output_id', portfolio_num, acc_num, loc_num, policy_num,
                'coverage_type_id']

        if include_loss_factor:
            group_by_cols = summary_cols + ['loss_factor_idx']
        else:
            group_by_cols = summary_cols
        guls_df = guls_df.loc[:, group_by_cols + ['loss_gul']]

        if not il and not ril:
            all_loss_cols = group_by_cols + ['loss_gul']
            all_losses_df = guls_df.loc[:, all_loss_cols]
            all_losses_df.drop_duplicates(keep=False, inplace=True)
        elif not ril:
            all_loss_cols = group_by_cols + ['loss_gul', 'loss_il']
            all_losses_df = all_losses_df.loc[:, all_loss_cols]
            summary_gul_df = pd.DataFrame(
                {'loss_gul': guls_df.groupby(group_by_cols)['loss_gul'].sum()}).reset_index()
            summary_il_df = pd.DataFrame(
                {'loss_il': all_losses_df.groupby(group_by_cols)['loss_il'].sum()}).reset_index()
            all_losses_df = summary_gul_df.merge(how='left', right=summary_il_df, on=group_by_cols)
        else:
            all_loss_cols = group_by_cols + ['loss_gul', 'loss_il', 'loss_ri']
            all_losses_df = all_losses_df.loc[:, all_loss_cols]
            summary_gul_df = pd.DataFrame(
                {'loss_gul': guls_df.groupby(group_by_cols)['loss_gul'].sum()}).reset_index()
            summary_il_df = pd.DataFrame(
                {'loss_il': all_losses_df.groupby(group_by_cols)['loss_il'].sum()}).reset_index()
            summary_ri_df = pd.DataFrame(
                {'loss_ri': all_losses_df.groupby(group_by_cols)['loss_ri'].sum()}).reset_index()
            all_losses_df = summary_gul_df.merge(how='left', right=summary_il_df, on=group_by_cols)
            all_losses_df = all_losses_df.merge(how='left', right=summary_ri_df, on=group_by_cols)

        for i in range(len(self.loss_factor)):

            if include_loss_factor:
                total_gul = guls_df[guls_df.loss_factor_idx == i].loss_gul.sum()
            else:
                total_gul = guls_df.loss_gul.sum()

            if not il and not ril:
                all_loss_cols = all_loss_cols + ['loss_gul']
                all_losses_df = guls_df.loc[:, all_loss_cols]
                all_losses_df.drop_duplicates(keep=False, inplace=True)
                header = \
                    'Losses (loss factor={:.2%}; total gul={:,.00f})'.format(
                        self.loss_factor[i],
                        total_gul)
            elif not ril:
                if include_loss_factor:
                    total_il = ils_df[ils_df.loss_factor_idx == i].loss_il.sum()
                else:
                    total_il = ils_df.loss_il.sum()

                header = \
                    'Losses (loss factor={:.2%}; total gul={:,.00f}; total il={:,.00f})'.format(
                        self.loss_factor[i],
                        total_gul, total_il)
            else:
                if include_loss_factor:
                    total_il = ils_df[ils_df.loss_factor_idx == i].loss_il.sum()
                    total_ri_net = rils_df[rils_df.loss_factor_idx == i].loss_ri.sum()
                else:
                    total_il = ils_df.loss_il.sum()
                    total_ri_net = rils_df.loss_ri.sum()
                total_ri_ceded = total_il - total_ri_net
                header = \
                    'Losses (loss factor={:.2%}; total gul={:,.00f}; total il={:,.00f}; total ri ceded={:,.00f})'.format(
                        self.loss_factor[i],
                        total_gul, total_il, total_ri_ceded)

            # Convert output cols to strings for formatting
            for c in group_by_cols:
                all_losses_df[c] = all_losses_df[c].apply(str)

            if self.print_summary:
                cols_to_print = all_loss_cols.copy()
                if False:
                    cols_to_print.remove('loss_factor_idx')
                if include_loss_factor:
                    print_dataframe(
                        all_losses_df[all_losses_df.loss_factor_idx == str(i)],
                        frame_header=header,
                        cols=cols_to_print)
                else:
                    print_dataframe(
                        all_losses_df,
                        frame_header=header,
                        cols=cols_to_print)

        if self.output_file:
            all_losses_df.to_csv(self.output_file, index=False, encoding='utf-8')

        if tmp_dir:
            tmp_dir.cleanup()

        return (il, ril)


class RunFmTest(ComputationStep):
    """
    Runs an FM test case and validates generated
    losses against expected losses.

    only use 'update_expected' for debugging
    it replaces the expected file with generated
    """

    step_params = [
        {'name': 'test_case_name', 'flag': '-c', 'type': str,
         'help': 'Runs a specific test sub-directory from "test_case_dir". If not set then run all tests found.'},
        {'name': 'list_tests', 'flag': '-l', 'action': 'store_true', 'help': 'List the valid test cases in the test directory rather than running'},
        {'name': 'test_case_dir', 'flag': '-t', 'default': os.getcwd(), 'is_path': True, 'pre_exist': True,
         'help': 'Test directory - should contain test directories containing OED files and expected results'},
        {'name': 'run_dir', 'flag': '-r', 'help': 'Run directory - where files should be generated. If not set temporary files will not be saved.'},
        {'name': 'num_subperils', 'flag': '-p', 'default': 1, 'type': int,
         'help': 'Set the number of subperils returned by deterministic key generator'},
        {'name': 'test_tolerance', 'type': float, 'help': 'Relative tolerance between expected values and results, default is "1e-4" or 0.0001',
         'default': 1e-4},
        {'name': 'fmpy', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'use fmcalc python version instead of c++ version'},
        {'name': 'fmpy_low_memory', 'default': False, 'type': str2bool, 'const': True, 'nargs': '?',
         'help': 'use memory map instead of RAM to store loss array (may decrease performance but reduce RAM usage drastically)'},
        {'name': 'fmpy_sort_output', 'default': True, 'type': str2bool, 'const': True, 'nargs': '?', 'help': 'order fmpy output by item_id'},
        {'name': 'update_expected', 'default': False},
        {'name': 'hashed_group_id', 'default': False},
        {'name': 'expected_output_dir', 'default': "expected"},
    ]

    def search_test_cases(self):
        case_names = []
        for test_case in os.listdir(path=self.test_case_dir):
            if os.path.exists(
                    os.path.join(self.test_case_dir, test_case, self.expected_output_dir)
            ):
                case_names.append(test_case)
        case_names.sort()
        return case_names, len(case_names)

    def _case_dir_is_valid_test(self):
        src_contents = [fn.lower() for fn in os.listdir(self.test_case_dir)]
        return 'location.csv' and 'account.csv' and 'expected' in src_contents

    def run(self):
        # Run test case given on CLI
        if self.test_case_name:
            return self.execute_test_case(self.test_case_name)

        # If 'test_case_dir' is a valid test run that dir directly
        if self._case_dir_is_valid_test():
            return self.execute_test_case('')

        # Search for valid cases in sub-dirs and run all found
        case_names, case_num = self.search_test_cases()

        # If '--list-tests' is selected print found cases and exit
        if self.list_tests:
            for name in case_names:
                self.logger.info(name)
            exit(0)

        if case_num < 1:
            raise OasisException(f'No vaild FM test cases found in "{self.test_case_dir}"')
        else:
            # If test_case not selected run all cases
            self.logger.info(f"Running: {case_num} Tests from '{self.test_case_dir}'")
            self.logger.info(f'Test names: {case_names}')
            failed_tests = []
            exit_status = 0
            for case in case_names:
                test_result = self.execute_test_case(case)

                if not test_result:
                    failed_tests.append(case)
                    exit_status = 1

            if len(failed_tests) == 0:
                self.logger.info("All tests passed")
            else:
                self.logger.info("{} test failed: ".format(len(failed_tests)))
                [self.logger.info(n) for n in failed_tests]
            exit(exit_status)

    def execute_test_case(self, test_case):
        if self.run_dir:
            tmp_dir = None
            run_dir = self.run_dir
        else:
            tmp_dir = tempfile.TemporaryDirectory()
            run_dir = tmp_dir.name

        test_dir = os.path.join(self.test_case_dir, test_case)
        output_level = 'loc'
        loss_factor_fp = os.path.join(test_dir, 'loss_factors.csv')
        loss_factor = []
        include_loss_factor = False

        if os.path.exists(loss_factor_fp):
            loss_factor = []
            include_loss_factor = True
            try:
                with open(loss_factor_fp, 'r') as csvfile:
                    reader = csv.DictReader(csvfile)
                    for row in reader:
                        loss_factor.append(
                            float(row['loss_factor']))
            except Exception as e:
                raise OasisException(f"Failed to read {loss_factor_fp}", e)
        else:
            loss_factor.append(1.0)

        output_file = os.path.join(run_dir, 'loc_summary.csv')
        (il, ril) = RunExposure(
            src_dir=test_dir,
            run_dir=run_dir,
            loss_factor=loss_factor,
            output_level=output_level,
            output_file=output_file,
            include_loss_factor=include_loss_factor,
            num_subperils=self.num_subperils,
            fmpy=self.fmpy,
            fmpy_low_memory=self.fmpy_low_memory,
            fmpy_sort_output=self.fmpy_sort_output,
            hashed_group_id=self.hashed_group_id,
        ).run()

        expected_data_dir = os.path.join(test_dir, self.expected_output_dir)

        if not os.path.exists(expected_data_dir):
            if self.update_expected:
                os.makedirs(expected_data_dir)
            else:
                raise OasisException(
                    'No subfolder named `expected` found in the input directory - '
                    'this subfolder should contain the expected set of GUL + IL '
                    'input files, optionally the RI input files, and the expected '
                    'set of GUL, IL and optionally the RI loss files'
                )

        files = ['keys.csv', 'loc_summary.csv']
        files += [
            '{}.csv'.format(fn)
            for ft, fn in chain(OASIS_FILES_PREFIXES['gul'].items(), OASIS_FILES_PREFIXES['il'].items())
        ]
        files += ['gul_summary_map.csv', 'guls.csv']
        if il:
            files += ['fm_summary_map.csv', 'ils.csv']
        if ril:
            files += ['rils.csv']

        test_result = True
        for f in files:
            generated = os.path.join(run_dir, f)
            expected = os.path.join(expected_data_dir, f)

            if not os.path.exists(expected):
                if self.update_expected:
                    shutil.copyfile(generated, expected)
                continue

            try:
                pd.testing.assert_frame_equal(
                    pd.read_csv(expected),
                    pd.read_csv(generated),
                    check_exact=False,
                    rtol=self.test_tolerance
                )
            except AssertionError:
                if self.update_expected:
                    shutil.copyfile(generated, expected)
                else:
                    print("Expected:")
                    with open(expected) as f:
                        self.logger.info(f.read())
                    print("Generated:")
                    with open(generated) as f:
                        self.logger.info(f.read())
                    raise OasisException(
                        f'\n FAIL: generated {generated} vs expected {expected}'
                    )
                    test_result = False
        if tmp_dir:
            tmp_dir.cleanup()
        return test_result



import numpy as np
from numpy.testing import assert_allclose
from .stream import event_agg_dtype, sidx_loss_dtype, EXTRA_VALUES
from .common import np_oasis_float
from .financial_structure import load_static


def stream_to_dict_array(stream_obj):
    stream_type = stream_obj.read(4)
    len_sample = np.frombuffer(stream_obj.read(4), dtype=np.int32)[0]

    buf = bytearray(8)
    mv = memoryview(buf)

    event_agg = np.ndarray(1, buffer=mv, dtype=event_agg_dtype)
    sidx_loss = np.ndarray(1, buffer=mv, dtype=sidx_loss_dtype)

    event_id_last = 0
    event_id, agg_id = 0, 0
    dict_array = {}
    while stream_obj.readinto(mv):
        if agg_id:
            sidx, loss = sidx_loss[0]
            if sidx == -3:
                sidx = 0
            elif sidx == -2:
                continue
            elif sidx == 0:
                agg_id = 0
                continue
            cur_array[sidx] = 0 if np.isnan(loss) else loss
        else:
            event_id, agg_id = event_agg[0]
            # if event_id_last != event_id:
            #     if event_id_last:
            #         break
            #     else:
            #         event_id_last = event_id
            cur_array = np.zeros(len_sample + EXTRA_VALUES, dtype=np_oasis_float)
            dict_array[(event_id, agg_id)] = cur_array

    return stream_type, len_sample, dict_array


def round_dict_array(dict_array, precision):
    for key, values in dict_array.items():
        values.round(decimals=precision, out=values)


def dict_array_to_np_array(dict_array, len_sample):
    res = np.empty(len(dict_array), dtype=np.dtype(f"i4, i4, ({len_sample + EXTRA_VALUES})f4"))
    for i, (event_id, agg_id) in enumerate(sorted(dict_array)):
        res[i] = event_id, agg_id, dict_array[(event_id, agg_id)]
    return res


def compare_streams(gul_stream, fm_stream_obj1, fm_stream_obj2, precision):
    fm_programme, fm_policytc, fm_profile, fm_xref, _, _ = load_static('./input')

    _, _, dict_array_gul = stream_to_dict_array(gul_stream)
    stream_type1, len_sample1, dict_array1 = stream_to_dict_array(fm_stream_obj1)
    stream_type2, len_sample2, dict_array2 = stream_to_dict_array(fm_stream_obj2)

    if stream_type1 != stream_type2:
        return f"stream have different type: {stream_type1}, {stream_type2}"

    if len_sample1 != len_sample2:
        return f"stream have different len_sample: {len_sample1}, {len_sample2}"

    keys1 = set(dict_array1)
    keys2 = set(dict_array2)

    missing_in_1 = keys2 - keys1
    missing_in_2 = keys1 - keys2

    if missing_in_1 or missing_in_2:
        msg = "some event_id, agg_id are not matching\n"
        for i, missing in enumerate([missing_in_1, missing_in_2]):
            if missing:
                msg += f"    {len(missing)} missing in {i+1} : {sorted(missing)[:10]}" \
                       f"{'...' if len(missing)>10 else ''}\n"
        return msg

    # round_dict_array(dict_array1, precision)
    # round_dict_array(dict_array2, precision)

    msg_list = []
    mismatch = 0
    for i, key in enumerate(dict_array1):
        try:
            assert_allclose(dict_array1[key], dict_array2[key], precision)
        except AssertionError:
            mismatch += 1
            msg_list.append(f"value mismatch for {key} index {i}:\n\t{dict_array_gul.get(key)}\n\t{dict_array1[key]}\n\t{dict_array2[key]}")
            output_id, agg_id, layer_id = np.extract(fm_xref['output_id'] == key[1], fm_xref)[0]
            cur_level, cur_agg_id = 1, agg_id
            while True:
                policytc = np.extract(np.logical_and(fm_policytc['level_id'] == cur_level,
                                                     fm_policytc['agg_id'] == agg_id,
                                                     np.logical_or(fm_policytc['layer_id'] == layer_id, fm_policytc['layer_id'] == 1)),
                                      fm_policytc)
                if policytc.shape[0] > 1:
                    policytc_id = np.extract(policytc['layer_id'] == layer_id, policytc['policytc_id'])
                    true_layer = layer_id
                else:
                    policytc_id = np.extract(policytc['layer_id'] == 1, policytc['policytc_id'])
                    true_layer = 1
                profile = np.extract(fm_profile['policytc_id'] == policytc_id, fm_profile)
                msg_list.append(str((cur_level, agg_id, true_layer, profile)))
                cur_level += 1
                parent = np.extract(np.logical_and(fm_programme['from_agg_id'] == agg_id,
                                                   fm_programme['level_id'] == cur_level),
                                    fm_programme)
                if not parent:
                    break
                else:
                    brothers = np.extract(np.logical_and(fm_programme['to_agg_id'] == parent['to_agg_id'],
                                                         fm_programme['level_id'] == cur_level),
                                          fm_programme)
                    msg_list.append(f"prothers {brothers}")

            if mismatch > 10:
                msg_list.append("...")
                break

    return "\n".join(msg_list)



"""
This file defines the functions that convert correlations.
"""
import os

from oasislmf.pytools.data_layer.oasis_files.correlations import CorrelationsData


def convert_csv_to_bin(file_path: str) -> None:
    """
    Converts the correlations data from a CSV file to a binary file.

    Args:
        file_path: (str) path to the file being converted

    Returns: None
    """
    data = CorrelationsData.from_csv(file_path=file_path)
    data.to_bin(file_path=file_path.replace("csv", "bin"))


def convert_bin_to_csv(file_path: str) -> None:
    """
    Converts the correlations data from a binary file to a CSV file.

    Args:
        file_path: (str) path to the file being converted

    Returns: None
    """
    data = CorrelationsData.from_bin(file_path=file_path)
    data.to_csv(file_path=file_path.replace("bin", "csv"))


def convert_csv_to_bin_main() -> None:
    """
    The command line entry point for converting CSV correlations file to a binary file.

    Returns: None
    """
    file_path = str(os.getcwd()) + "/correlations.csv"
    convert_csv_to_bin(file_path=file_path)


def convert_bin_to_csv_main() -> None:
    """
    The command line entry point for converting binary correlations file to a CSV file.

    Returns: None
    """
    file_path = str(os.getcwd()) + "/correlations.bin"
    convert_bin_to_csv(file_path=file_path)



__all__ = [
    'get_calc_rules',
    'get_step_calc_rules'
]

import os

from .data import get_dataframe
from .defaults import (
    STATIC_DATA_FP,
)


# Ktools calc. rules
def get_calc_rules(policy_layer=False):
    calc_rules_filename = 'calc_rules_direct'
    if policy_layer:
        calc_rules_filename += '_layer'
    calc_rules_filename += '.csv'
    fp = os.path.join(STATIC_DATA_FP, calc_rules_filename)
    return get_dataframe(src_fp=fp)


def get_step_calc_rules():
    """
    Get dataframe of calc rules for step policies

    :return: dataframe of calc rules for step policies
    :rtype: pandas.DataFrame
    """
    fp = os.path.join(STATIC_DATA_FP, 'calc_rules_step.csv')
    return get_dataframe(src_fp=fp)



"""
This file defines the loading and saving for correlations data.
"""
from typing import Optional

import numpy as np
import pandas as pd

from oasislmf.pytools.getmodel.common import Correlation


class CorrelationsData:
    """
    This class is responsible for managing the loading and saving of correlation data from binary and CSV files.

    Attributes:
        data (Optional[pd.DataFrame): correlation data that is either loaded or saved
    """
    COLUMNS = ["item_id", "peril_correlation_group", "correlation_value"]

    def __init__(self, data: Optional[pd.DataFrame] = None) -> None:
        """
        The constructor for the CorrelationsData class.

        Args:
            data: (Optional[pd.DataFrame] default is None but if supplied must have the following columns:
                                          [item_id, peril_correlation_group, correlation_value]
        """
        self.data: Optional[pd.DataFrame] = data

    @staticmethod
    def from_csv(file_path: str) -> "CorrelationsData":
        """
        Loads correlations data from a CSV file.

        Args:
            file_path: (str) the path to the CSV file housing the data

        Returns: (CorrelationsData) the loaded data from the CSV file
        """
        return CorrelationsData(data=pd.read_csv(file_path))

    @staticmethod
    def from_bin(file_path: str) -> "CorrelationsData":
        """
        Loads correlations data from a binary file.

        Args:
            file_path: (str) the path to the binary file housing the data

        Returns: (CorrelationsData) the loaded data from the binary file
        """
        data = pd.DataFrame(np.fromfile(file_path, dtype=Correlation))
        data["item_id"] = list(range(1, len(data) + 1))
        data = data[["item_id", "peril_correlation_group", "correlation_value"]]
        return CorrelationsData(data=data)

    def to_csv(self, file_path: str) -> None:
        """
        Writes self.data to a CSV file.

        Args:
            file_path: (str) the file path for the CSV file to be written to

        Returns: None
        """
        self.data.to_csv(file_path, index=False)

    def to_bin(self, file_path: str) -> None:
        """
        Writes self.data to a binary file.

        Args:
            file_path: (str) the file path for the binary file to be written to

        Returns: None
        """
        data = np.array(list(self.data.drop("item_id", axis=1).itertuples(index=False)), dtype=Correlation)
        data.tofile(file_path)



"""
This file contains specific functionality needed for aggregate vulnerabilities.

"""
import logging
import os

import numba as nb
import numpy as np
import pandas as pd
from numba import njit
from numba.typed import Dict, List
from numba.types import int32 as nb_int32
from numba.types import uint32 as nb_uint32

from oasislmf.pytools.getmodel.common import areaperil_int
from oasislmf.pytools.gulmc.common import (AggregateVulnerability,
                                           VulnerabilityWeight)

AGG_VULN_WEIGHTS_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
AGG_VULN_WEIGHTS_VAL_TYPE = nb.types.int32

logger = logging.getLogger(__name__)


@njit(cache=True)
def gen_empty_agg_vuln_to_vuln_ids():
    """Generate empty map to store the definitions of aggregate vulnerability functions.

    Returns:
        dict[int, list[int]]: map of aggregate vulnerability id to list of vulnerability ids.
    """
    return Dict.empty(nb_int32, List.empty_list(nb_int32))


@njit(cache=True)
def gen_empty_areaperil_vuln_ids_to_weights():
    """Generate empty map to store the weights of individual vulnerability functions in each aggregate vulnerability.

    Returns:
        dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]: map of areaperil_id, vulnerability id to weight.
    """
    return Dict.empty(AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE)


def read_aggregate_vulnerability(path, ignore_file_type=set()):
    """Load the aggregate vulnerability definitions from file.

    Args:
        path (str): the path pointing to the file
        ignore_file_type (Set[str]): file extension to ignore when loading.

    Returns:
        np.array[AggregateVulnerability]: aggregate vulnerability table.
    """
    input_files = set(os.listdir(path))

    if "aggregate_vulnerability.bin" in input_files and "bin" not in ignore_file_type:
        fname = os.path.join(path, 'aggregate_vulnerability.bin')
        logger.debug(f"loading {fname}")
        aggregate_vulnerability = np.memmap(fname, dtype=AggregateVulnerability, mode='r')

    elif "aggregate_vulnerability.csv" in input_files and "csv" not in ignore_file_type:
        fname = os.path.join(path, 'aggregate_vulnerability.csv')
        logger.debug(f"loading {fname}")
        aggregate_vulnerability = np.loadtxt(fname, dtype=AggregateVulnerability, delimiter=",", skiprows=1, ndmin=1)

    else:
        aggregate_vulnerability = None
        logging.warning('Aggregate vulnerability table not found at {path}. Continuing without aggregate vulnerability definitions.')

    return aggregate_vulnerability


def read_vulnerability_weights(path, ignore_file_type=set()):
    """Load the vulnerability weights definitions from file.

    Args:
        path (str): the path pointing to the file
        ignore_file_type (Set[str]): file extension to ignore when loading.

    Returns:
        np.array[VulnerabilityWeight]: vulnerability weights table.
    """
    input_files = set(os.listdir(path))

    if "weights.bin" in input_files and "bin" not in ignore_file_type:
        fname = os.path.join(path, 'weights.bin')
        logger.debug(f"loading {fname}")
        aggregate_weights = np.memmap(fname, dtype=VulnerabilityWeight, mode='r')

    elif "weights.csv" in input_files and "csv" not in ignore_file_type:
        fname = os.path.join(path, 'weights.csv')
        logger.debug(f"loading {fname}")
        aggregate_weights = np.loadtxt(fname, dtype=VulnerabilityWeight, delimiter=",", skiprows=1, ndmin=1)

    else:
        aggregate_weights = None
        logging.warning('Vulnerability weights not found at {path}. Continuing without vulnerability weights definitions.')

    return aggregate_weights


def process_aggregate_vulnerability(aggregate_vulnerability):
    """Rearrange aggregate vulnerability definitions from tabular format to a map between aggregate
    vulnerability id and the list of vulnerability ids that it is made of.

    Args:
        aggregate_vulnerability (np.array[AggregateVulnerability]): aggregate vulnerability table.

    Returns:
        dict[int, list[int]]: map of aggregate vulnerability id to list of vulnerability ids.
    """
    agg_vuln_to_vuln_id = gen_empty_agg_vuln_to_vuln_ids()

    if aggregate_vulnerability is not None:

        agg_vuln_df = pd.DataFrame(aggregate_vulnerability)
        # init agg_vuln_to_vuln_id to allow numba to compile later functions
        # vulnerability_id and aggregate_vulnerability_id are remapped to the internal ids
        # using the vulnd_dict map that contains only the vulnerability_id used in this portfolio.

        # here we read all aggregate vulnerability_id, then, after processing the items file,
        # we will filter out the aggregate vulnerability that are not used in this portfolio.
        for agg, grp in agg_vuln_df.groupby('aggregate_vulnerability_id'):
            agg_vuln_id = nb_int32(agg)

            if agg_vuln_id not in agg_vuln_to_vuln_id:
                agg_vuln_to_vuln_id[agg_vuln_id] = List.empty_list(nb_int32)

            for entry in grp['vulnerability_id'].to_list():
                agg_vuln_to_vuln_id[agg_vuln_id].append(nb_int32(entry))

    return agg_vuln_to_vuln_id


def process_vulnerability_weights(aggregate_weights, agg_vuln_to_vuln_id):
    """Rearrange vulnerability weights from tabular format to a map between (areaperil_id, vulnerability_id) and the vulnerability weight.

    Args:
        aggregate_weights (np.array[VulnerabilityWeight]): vulnerability weights table.
        agg_vuln_to_vuln_id (dict[int, list[int]]): map of aggregate vulnerability id to list of vulnerability ids.

    Returns:
        dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]: map of areaperil_id, vulnerability id to weight.
    """
    areaperil_vuln_id_to_weight = gen_empty_areaperil_vuln_ids_to_weights()

    if aggregate_weights is not None:
        agg_weight_df = pd.DataFrame(aggregate_weights)

        if len(agg_vuln_to_vuln_id) > 0:
            # at least one aggregate vulnerability is defined
            for agg, grp in agg_weight_df.groupby(['areaperil_id', 'vulnerability_id']):
                areaperil_vuln_id_to_weight[(nb_uint32(agg[0]), nb_int32(agg[1]))] = nb_int32(grp['weight'].to_list()[0])

    return areaperil_vuln_id_to_weight


@njit(cache=True)
def map_agg_vuln_ids_to_agg_vuln_idxs(agg_vulns, agg_vuln_to_vuln_id, vuln_dict):
    """For each aggregate vulnerability listed in `agg_vulns`, map the individual vulnerability_ids that compose it
    to the indices where they are stored in `vuln_array`.

    Args:
        agg_vulns (List[int32])
        agg_vuln_to_vuln_id (dict[int, list[int]]): map of aggregate vulnerability id to list of vulnerability ids.
        vuln_dict (Tuple[Dict[int, int]): vulnerability dictionary.

    Returns:
        dict[int, list[int]]: map between aggregate vulnerability id and the list of indices where the individual vulnerability_ids
          that compose it are stored in `vuln_array`.
    """
    agg_vuln_to_vuln_idxs = Dict.empty(nb_int32, List.empty_list(nb_int32))

    for agg in agg_vulns:
        agg_vuln_to_vuln_idxs[agg] = List([vuln_dict[vuln] for vuln in agg_vuln_to_vuln_id[agg]])

    return agg_vuln_to_vuln_idxs


@njit(cache=True)
def map_areaperil_vuln_id_to_weight_to_areaperil_vuln_idx_to_weight(areaperil_to_vulns, areaperil_vuln_id_to_weight, vuln_dict):
    """Make map between aggregate vulnerability id and the list of sub-vulnerability ids of which they are composed, where the
    value of the sub-vulnerability id is the internal pointer (i.e., the index, startinf from 0) to the dense array where they are stored,
    not the vulnerability id (which starts from 1).

    Args:
        areaperil_to_vulns (List[int32])
        areaperil_vuln_id_to_weight (dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]): map of areaperil_id, vulnerability id to weight.
        vuln_dict (Tuple[Dict[int, int]): vulnerability dictionary.

    Returns:
        dict[AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE]: map between the areaperil id and the index where the vulnerability function 
          is stored in `vuln_array` and the vulnerability weight.
    """
    areaperil_vuln_idx_to_weight = Dict.empty(AGG_VULN_WEIGHTS_KEY_TYPE, AGG_VULN_WEIGHTS_VAL_TYPE)

    if len(areaperil_vuln_id_to_weight) > 0:
        for ap in areaperil_to_vulns:
            for vuln in areaperil_to_vulns[ap]:
                if tuple((ap, vuln)) in areaperil_vuln_id_to_weight:
                    # if the weights file contains this definition, add this entry to the map.
                    # otherwise, the weight of this (areaperil_id, vulnerability_id) is assumed to be zero during loss calculation.
                    areaperil_vuln_idx_to_weight[tuple((ap, vuln_dict[vuln]))] = areaperil_vuln_id_to_weight[tuple((ap, vuln))]

    return areaperil_vuln_idx_to_weight



from .policy import calc
from .policy_extras import calc as calc_extra
from .common import np_oasis_float, np_oasis_int, EXTRA_VALUES, null_index

from numba import njit
import numpy as np
import os
import logging
logger = logging.getLogger(__name__)

DEDUCTIBLE = 0
OVERLIMIT = 1
UNDERLIMIT = 2


@njit(cache=True)
def get_base_children(node, children, nodes_array, temp_children_queue):
    len_children = children[node['children']]
    if len_children:
        temp_children_queue[:len_children] = children[node['children'] + 1: node['children'] + len_children + 1]
        temp_children_queue[len_children] = null_index
        queue_end = len_children
        i = 0
        child_i = 0
        while temp_children_queue[i] != null_index:
            parent = nodes_array[temp_children_queue[i]]
            len_children = children[parent['children']]
            if len_children:
                temp_children_queue[queue_end: queue_end + len_children] = children[parent['children'] + 1: parent['children'] + len_children + 1]
                queue_end += len_children
                temp_children_queue[queue_end] = null_index
            else:
                temp_children_queue[child_i] = temp_children_queue[i]
                child_i += 1
            i += 1
    else:
        temp_children_queue[0] = node['node_id']
        child_i = 1
    return child_i


@njit(cache=True)
def first_time_layer(profile_len, base_children_len, temp_children_queue, nodes_array,
                     sidx_indptr, sidx_indexes,
                     loss_indptr, loss_val, loss_ptr_i,
                     ):
    """
    first time there is a back allocation with multiple layer, we duplicate loss and extra from layer 1 to the other layers
    """
    for base_child_i in range(base_children_len):
        child = nodes_array[temp_children_queue[base_child_i]]
        child_val_len = sidx_indptr[sidx_indexes[child['node_id']] + 1] - sidx_indptr[sidx_indexes[child['node_id']]]
        child_loss_val_layer_0 = loss_val[loss_indptr[child['loss']]:
                                          loss_indptr[child['loss']] + child_val_len]
        for p in range(1, profile_len):
            loss_indptr[child['loss'] + p] = loss_ptr_i
            loss_val[loss_ptr_i: loss_ptr_i + child_val_len] = child_loss_val_layer_0
            loss_ptr_i += child_val_len
    return loss_ptr_i


@njit(cache=True)
def first_time_layer_extra(profile_len, base_children_len, temp_children_queue, nodes_array,
                           sidx_indptr, sidx_indexes,
                           loss_indptr, loss_val, loss_ptr_i,
                           extras_indptr, extras_val, extras_ptr_i
                           ):
    """
    first time there is a back allocation with multiple layer, we duplicate loss and extra from layer 1 to the other layers
    """
    for base_child_i in range(base_children_len):
        child = nodes_array[temp_children_queue[base_child_i]]
        child_val_len = sidx_indptr[sidx_indexes[child['node_id']] + 1] - sidx_indptr[sidx_indexes[child['node_id']]]
        child_loss_val_layer_0 = loss_val[loss_indptr[child['loss']]:
                                          loss_indptr[child['loss']] + child_val_len]
        child_extra_val_layer_0 = extras_val[extras_indptr[child['extra']]:
                                             extras_indptr[child['extra']] + child_val_len]
        for p in range(1, profile_len):
            loss_indptr[child['loss'] + p] = loss_ptr_i
            loss_val[loss_ptr_i: loss_ptr_i + child_val_len] = child_loss_val_layer_0
            loss_ptr_i += child_val_len

            extras_indptr[child['extra'] + p] = extras_ptr_i
            extras_val[extras_ptr_i: extras_ptr_i + child_val_len] = child_extra_val_layer_0
            extras_ptr_i += child_val_len

    return loss_ptr_i, extras_ptr_i


@njit(cache=True, fastmath=True)
def aggregate_children_extras(node, len_children, nodes_array, children, temp_children_queue,
                              temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, sidx_ptr_i, sidx_i, all_sidx,
                              temp_node_loss, loss_indptr, loss_val, loss_ptr_i,
                              temp_node_extras, extras_indptr, extras_val, extras_ptr_i):
    sidx_created = False
    node_sidx_start = sidx_ptr_i
    node_sidx_end = 0
    sidx_indexes[node['node_id']], sidx_i = sidx_i, sidx_i + 1

    for p in range(node['profile_len']):
        p_temp_node_loss = temp_node_loss[p]
        p_temp_node_extras = temp_node_extras[p]

        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            child = nodes_array[children[c]]
            child_sidx_val = sidx_val[sidx_indptr[sidx_indexes[child['node_id']]]:
                                      sidx_indptr[sidx_indexes[child['node_id']] + 1]]

            if p == 1 and loss_indptr[child['loss'] + p] == loss_indptr[child['loss']]:
                # this is the first time child branch has multiple layer we create views for root children
                base_children_len = get_base_children(child, children, nodes_array, temp_children_queue)
                # print('new layers', child['level_id'], child['agg_id'], node['profile_len'])
                loss_ptr_i, extras_ptr_i = first_time_layer_extra(node['profile_len'], base_children_len, temp_children_queue, nodes_array,
                                                                  sidx_indptr, sidx_indexes,
                                                                  loss_indptr, loss_val, loss_ptr_i,
                                                                  extras_indptr, extras_val, extras_ptr_i
                                                                  )

            child_loss = loss_val[loss_indptr[child['loss'] + p]:
                                  loss_indptr[child['loss'] + p] + child_sidx_val.shape[0]]
            child_extra = extras_val[extras_indptr[child['extra'] + p]:
                                     extras_indptr[child['extra'] + p] + child_sidx_val.shape[0]]
            # print('child', child['level_id'], child['agg_id'], p, loss_indptr[child['loss'] + p], child_loss[0], child['extra'], extras_indptr[child['extra'] + p])
            for indptr in range(child_sidx_val.shape[0]):
                temp_node_sidx[child_sidx_val[indptr]] = True
                p_temp_node_loss[child_sidx_val[indptr]] += child_loss[indptr]
                p_temp_node_extras[child_sidx_val[indptr]] += child_extra[indptr]
        # print('res', p, p_temp_node_loss[-3], p_temp_node_extras[-3])

        loss_indptr[node['loss'] + p] = loss_ptr_i
        extras_indptr[node['extra'] + p] = extras_ptr_i
        if sidx_created:
            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                loss_val[loss_ptr_i] = p_temp_node_loss[sidx_val[node_sidx_cur]]
                loss_ptr_i += 1
                extras_val[extras_ptr_i] = p_temp_node_extras[sidx_val[node_sidx_cur]]
                extras_ptr_i += 1

        else:
            for sidx in all_sidx:
                if temp_node_sidx[sidx]:
                    sidx_val[sidx_ptr_i] = sidx
                    sidx_ptr_i += 1

                    loss_val[loss_ptr_i] = p_temp_node_loss[sidx]
                    loss_ptr_i += 1

                    extras_val[extras_ptr_i] = p_temp_node_extras[sidx]
                    extras_ptr_i += 1

            node_sidx_end = sidx_ptr_i
            node_val_len = node_sidx_end - node_sidx_start
            sidx_indptr[sidx_i] = sidx_ptr_i
            sidx_created = True
    # print('node', node['node_id'], node['agg_id'], loss_indptr[node['loss']], temp_node_loss[:node['profile_len'], -3], temp_node_extras[:node['profile_len'], -3])
    # fill up all layer if necessary
    for layer in range(node['profile_len'], node['layer_len']):
        loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]
        extras_indptr[node['extra'] + layer] = extras_indptr[node['extra']]

    return node_val_len, sidx_i, sidx_ptr_i, loss_ptr_i, extras_ptr_i


@njit(cache=True, fastmath=True)
def aggregate_children(node, len_children, nodes_array, children, temp_children_queue,
                       temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, sidx_ptr_i, sidx_i, all_sidx,
                       temp_node_loss, loss_indptr, loss_val, loss_ptr_i):
    sidx_created = False
    node_sidx_start = sidx_ptr_i
    node_sidx_end = 0
    sidx_indexes[node['node_id']], sidx_i = sidx_i, sidx_i + 1
    for p in range(node['profile_len']):
        p_temp_node_loss = temp_node_loss[p]
        for c in range(node['children'] + 1, node['children'] + len_children + 1):
            child = nodes_array[children[c]]
            child_sidx_val = sidx_val[sidx_indptr[sidx_indexes[child['node_id']]]:
                                      sidx_indptr[sidx_indexes[child['node_id']] + 1]]
            if p == 1 and loss_indptr[child['loss'] + p] == loss_indptr[child['loss']]:
                # this is the first time child branch has multiple layer we create views for root children
                base_children_len = get_base_children(child, children, nodes_array, temp_children_queue)
                loss_ptr_i = first_time_layer(node['profile_len'], base_children_len, temp_children_queue, nodes_array,
                                              sidx_indptr, sidx_indexes,
                                              loss_indptr, loss_val, loss_ptr_i
                                              )
            child_loss = loss_val[loss_indptr[child['loss'] + p]:
                                  loss_indptr[child['loss'] + p] + child_sidx_val.shape[0]]

            for indptr in range(child_sidx_val.shape[0]):
                temp_node_sidx[child_sidx_val[indptr]] = True
                p_temp_node_loss[child_sidx_val[indptr]] += child_loss[indptr]

        loss_indptr[node['loss'] + p] = loss_ptr_i
        if sidx_created:
            for node_sidx_cur in range(node_sidx_start, node_sidx_end):
                loss_val[loss_ptr_i] = p_temp_node_loss[sidx_val[node_sidx_cur]]
                loss_ptr_i += 1

        else:
            for sidx in all_sidx:
                if temp_node_sidx[sidx]:
                    sidx_val[sidx_ptr_i] = sidx
                    sidx_ptr_i += 1
                    temp_node_sidx[sidx] = False

                    loss_val[loss_ptr_i] = p_temp_node_loss[sidx]
                    loss_ptr_i += 1

            node_sidx_end = sidx_ptr_i
            node_val_len = node_sidx_end - node_sidx_start
            sidx_indptr[sidx_i] = sidx_ptr_i
            sidx_created = True

    # fill up all layer if necessary
    for layer in range(node['profile_len'], node['layer_len']):
        loss_indptr[node['loss'] + layer] = loss_indptr[node['loss']]

    return node_val_len, sidx_i, sidx_ptr_i, loss_ptr_i


@njit(cache=True)
def set_parent_next_compute(parent_id, child_id, nodes_array, children, next_compute_i, computes):
    parent = nodes_array[parent_id]
    parent_children_len = children[parent['children']] + 1
    children[parent['children']] = parent_children_len
    children[parent['children'] + parent_children_len] = child_id
    if parent_children_len == 1:  # first time parent is seen
        computes[next_compute_i] = parent_id
        return next_compute_i + 1
    else:
        return next_compute_i


@njit(cache=True, fastmath=True)
def back_alloc_extra_a2(base_children_len, temp_children_queue, nodes_array, p,
                        node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                        loss_in, loss_out, temp_node_loss, loss_indptr, loss_val,
                        extra, temp_node_extras, extras_indptr, extras_val):
    if base_children_len == 1:  # this is a base children, we only need to assign loss_in
        loss_in[:] = loss_out
    else:
        # back allocation rules:
        # if deductible grows, deductible and loss are allocated based on loss
        # else it means it is reallocated to loss,
        #   if underlimit is still >0 then extra loss and deductible are allocated based on underlimit
        #   else                           extra loss and deductible are allocated based on deductible
        # if overlimit grows, more loss is over limit so it is reallocated based on loss
        # else it is reallocated based on overlimit
        # if underlimit grows, more loss has been deducted so we reallocate based on loss
        # else it is reallocated based on underlimit

        for i in range(node_val_len):
            diff = extra[i, DEDUCTIBLE] - temp_node_extras[p, node_sidx[i], DEDUCTIBLE]
            if diff >= 0:
                realloc = 0
                if loss_in[i] > 0:
                    temp_node_extras[p, node_sidx[i], DEDUCTIBLE] = diff / loss_in[i]
                    temp_node_loss[p, node_sidx[i]] = loss_out[i] / loss_in[i]
                else:
                    temp_node_extras[p, node_sidx[i], DEDUCTIBLE] = 0
                    temp_node_loss[p, node_sidx[i]] = 0
            else:
                realloc = diff  # to loss or to over

                if extra[i, UNDERLIMIT] > 0:
                    temp_node_extras[p, node_sidx[i], DEDUCTIBLE] = diff / temp_node_extras[p, node_sidx[i], UNDERLIMIT]
                else:
                    temp_node_extras[p, node_sidx[i], DEDUCTIBLE] = diff / temp_node_extras[p, node_sidx[i], DEDUCTIBLE]
                temp_node_loss[p, node_sidx[i]] = loss_out[i] / (loss_in[i] - diff)

            diff = extra[i, OVERLIMIT] - temp_node_extras[p, node_sidx[i], OVERLIMIT]
            if diff > 0:
                temp_node_extras[p, node_sidx[i], OVERLIMIT] = diff / (loss_in[i] - realloc)
            elif diff == 0:
                temp_node_extras[p, node_sidx[i], OVERLIMIT] = 0
            else:  # we set it to <0 to be able to check it later
                temp_node_extras[p, node_sidx[i], OVERLIMIT] = - extra[i, OVERLIMIT] / temp_node_extras[
                    p, node_sidx[i], OVERLIMIT]

            diff = extra[i, UNDERLIMIT] - temp_node_extras[p, node_sidx[i], UNDERLIMIT]
            if diff > 0:
                temp_node_extras[p, node_sidx[i], UNDERLIMIT] = diff / loss_in[i]
            elif diff == 0:
                temp_node_extras[p, node_sidx[i], UNDERLIMIT] = 0
            else:  # we set it to <0 to be able to check it later
                temp_node_extras[p, node_sidx[i], UNDERLIMIT] = - extra[i, UNDERLIMIT] / temp_node_extras[
                    p, node_sidx[i], UNDERLIMIT]

            loss_in[i] = loss_out[i]

        for base_child_i in range(base_children_len):
            child = nodes_array[temp_children_queue[base_child_i]]

            child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
            child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
            child_val_len = child_sidx_end - child_sidx_start

            child_sidx = sidx_val[child_sidx_start: child_sidx_end]
            child_loss = loss_val[loss_indptr[child['loss'] + p]: loss_indptr[child['loss'] + p] + child_val_len]
            child_extra = extras_val[
                extras_indptr[child['extra'] + p]: extras_indptr[child['extra'] + p] + child_val_len]

            for i in range(child_val_len):
                if temp_node_extras[p, child_sidx[i], DEDUCTIBLE] < 0:  # realloc loss
                    if temp_node_extras[p, child_sidx[i], UNDERLIMIT] == 0:
                        realloc = temp_node_extras[p, child_sidx[i], DEDUCTIBLE] * child_extra[i, DEDUCTIBLE]
                    else:
                        realloc = temp_node_extras[p, child_sidx[i], DEDUCTIBLE] * child_extra[i, UNDERLIMIT]
                    if temp_node_extras[p, child_sidx[i], OVERLIMIT] >= 0:
                        child_extra[i, OVERLIMIT] = child_extra[i, OVERLIMIT] + temp_node_extras[
                            p, child_sidx[i], OVERLIMIT] * (child_loss[i] - realloc)
                    else:
                        child_extra[i, OVERLIMIT] = - temp_node_extras[p, child_sidx[i], OVERLIMIT] * child_extra[
                            i, OVERLIMIT]

                    child_loss[i] = (child_loss[i] - realloc) * temp_node_loss[p, child_sidx[i]]
                    child_extra[i, DEDUCTIBLE] = child_extra[i, DEDUCTIBLE] + realloc
                    child_extra[i, UNDERLIMIT] = - temp_node_extras[p, child_sidx[i], UNDERLIMIT] * child_extra[
                        i, UNDERLIMIT]

                else:
                    if temp_node_extras[p, child_sidx[i], OVERLIMIT] >= 0:
                        child_extra[i, OVERLIMIT] = child_extra[i, OVERLIMIT] + temp_node_extras[
                            p, child_sidx[i], OVERLIMIT] * child_loss[i]
                    else:
                        child_extra[i, OVERLIMIT] = - temp_node_extras[p, child_sidx[i], OVERLIMIT] * child_extra[
                            i, OVERLIMIT]

                    if temp_node_extras[p, child_sidx[i], UNDERLIMIT] >= 0:
                        child_extra[i, UNDERLIMIT] = child_extra[i, UNDERLIMIT] + temp_node_extras[
                            p, child_sidx[i], UNDERLIMIT] * child_loss[i]
                    else:
                        child_extra[i, UNDERLIMIT] = - temp_node_extras[p, child_sidx[i], UNDERLIMIT] * child_extra[
                            i, UNDERLIMIT]

                    child_extra[i, DEDUCTIBLE] = child_extra[i, DEDUCTIBLE] + temp_node_extras[
                        p, child_sidx[i], DEDUCTIBLE] * child_loss[i]
                    child_loss[i] = child_loss[i] * temp_node_loss[p, child_sidx[i]]
            # print('ba', child['level_id'], child['agg_id'], p, loss_indptr[child['loss'] + p], child_loss[0], temp_node_loss[p, -3], extras_indptr[child['extra'] + p], child_extra[0])


@njit(cache=True, fastmath=True)
def back_alloc_a2(base_children_len, temp_children_queue, nodes_array, p,
                  node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                  loss_in, loss_out, temp_node_loss, loss_indptr, loss_val):
    if base_children_len == 1:
        loss_in[:] = loss_out
    else:
        for i in range(node_val_len):
            if loss_out[i]:
                temp_node_loss[p, node_sidx[i]] = loss_out[i] / loss_in[i]
            else:
                temp_node_loss[p, node_sidx[i]] = 0
            loss_in[i] = loss_out[i]

        for base_child_i in range(base_children_len):
            child = nodes_array[temp_children_queue[base_child_i]]

            child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
            child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
            child_val_len = child_sidx_end - child_sidx_start

            child_sidx = sidx_val[child_sidx_start: child_sidx_end]
            child_loss = loss_val[loss_indptr[child['loss'] + p]: loss_indptr[child['loss'] + p] + child_val_len]

            for i in range(child_val_len):
                child_loss[i] = child_loss[i] * temp_node_loss[p, child_sidx[i]]
            # print('ba', child['level_id'], child['agg_id'], p, loss_indptr[child['loss'] + p], child_loss[0], temp_node_loss[p, -3])


@njit(cache=True, fastmath=True)
def compute_event(compute_info,
                  net_loss,
                  nodes_array,
                  node_parents_array,
                  node_profiles_array,
                  len_array, max_sidx_val, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val, extras_indptr, extras_val,
                  children,
                  computes,
                  next_compute_i,
                  item_parent_i,
                  fm_profile,
                  stepped):
    # len_array = computation_structure.len_array
    # max_sidx_val = computation_structure.max_sidx_val
    # sidx_indexes = computation_structure.sidx_indexes
    # sidx_indptr = computation_structure.sidx_indptr
    # sidx_val = computation_structure.sidx_val
    # loss_indptr = computation_structure.loss_indptr
    # loss_val = computation_structure.loss_val
    # extras_indptr = computation_structure.extras_indptr
    # extras_val = computation_structure.extras_val

    sidx_ptr_i = sidx_indptr[next_compute_i]
    loss_ptr_i = sidx_ptr_i
    extras_ptr_i = 0
    sidx_i = next_compute_i
    compute_i = 0

    temp_node_sidx = np.zeros(len_array, dtype=np_oasis_int)
    temp_node_loss_sparse = np.zeros(len_array, dtype=np_oasis_float)
    temp_node_loss = np.zeros((compute_info['max_layer'], len_array), dtype=np.float64)
    temp_node_extras = np.zeros((compute_info['max_layer'], len_array, 3), dtype=np_oasis_float)
    temp_children_queue = np.empty(nodes_array.shape[0], dtype=np_oasis_int)

    # create all sidx array
    all_sidx = np.empty(max_sidx_val + EXTRA_VALUES, dtype=np_oasis_int)
    all_sidx[0] = -5
    all_sidx[1] = -3
    all_sidx[2] = -1
    all_sidx[3:] = np.arange(1, max_sidx_val + 1)

    is_net_loss_or_allocation_rule_a1 = net_loss or (compute_info['allocation_rule'] == 1)
    is_allocation_rule_a0 = compute_info['allocation_rule'] == 0
    is_allocation_rule_a1 = compute_info['allocation_rule'] == 1
    is_allocation_rule_a2 = compute_info['allocation_rule'] == 2

    for level in range(compute_info['start_level'], compute_info['max_level'] + 1):  # perform the bottom up loss computation
        # print(level, next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i+2], computes[next_compute_i-2: next_compute_i+1])

        next_compute_i += 1  # next_compute_i will stay at 0 wich will stop the while loop and start the next level
        level_start_compute_i = compute_i

        while computes[compute_i]:
            compute_node, compute_i = nodes_array[computes[compute_i]], compute_i + 1
            # compute loss sums and extra sum
            len_children = children[compute_node['children']]

            # step 1 depending on the number of children, we aggregate of copy the loss of the previous level
            # we fill up correctly temp_node_loss and temp_node_extras
            if len_children:  # if no children and in compute then layer 1 loss is already set
                if len_children > 1:
                    storage_node = compute_node
                    temp_node_loss.fill(0)
                    if storage_node['extra'] == null_index:
                        (node_val_len,
                         sidx_i,
                         sidx_ptr_i,
                         loss_ptr_i) = aggregate_children(storage_node, len_children, nodes_array, children, temp_children_queue,
                                                          temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, sidx_ptr_i,
                                                          sidx_i, all_sidx,
                                                          temp_node_loss, loss_indptr, loss_val, loss_ptr_i)
                    else:
                        temp_node_extras.fill(0)
                        (node_val_len,
                         sidx_i,
                         sidx_ptr_i,
                         loss_ptr_i,
                         extras_ptr_i) = aggregate_children_extras(storage_node, len_children, nodes_array, children, temp_children_queue,
                                                                   temp_node_sidx, sidx_indexes, sidx_indptr, sidx_val, sidx_ptr_i,
                                                                   sidx_i, all_sidx,
                                                                   temp_node_loss, loss_indptr, loss_val, loss_ptr_i,
                                                                   temp_node_extras, extras_indptr, extras_val, extras_ptr_i)
                    node_sidx = sidx_val[sidx_ptr_i - node_val_len: sidx_ptr_i]

                else:  # only 1 children
                    storage_node = nodes_array[children[compute_node['children'] + 1]]
                    # positive sidx are the same as child
                    node_sidx = sidx_val[sidx_indptr[sidx_indexes[storage_node['node_id']]]:sidx_indptr[sidx_indexes[storage_node['node_id']] + 1]]
                    node_val_len = node_sidx.shape[0]

                    if compute_node['profile_len'] > 1 and loss_indptr[storage_node['loss'] + 1] == loss_indptr[storage_node['loss']]:
                        # first time layer, we need to create view for storage_node and copy loss and extra
                        node_loss = loss_val[loss_indptr[storage_node['loss']]:loss_indptr[storage_node['loss']] + node_val_len]

                        for p in range(1, compute_node['profile_len']):
                            loss_indptr[storage_node['loss'] + p] = loss_ptr_i
                            loss_val[loss_ptr_i: loss_ptr_i + node_val_len] = node_loss
                            loss_ptr_i += node_val_len

                        if compute_node['extra'] != null_index:
                            node_extras = extras_val[extras_indptr[storage_node['extra']]:extras_indptr[storage_node['extra']] + node_val_len]
                            for p in range(1, compute_node['profile_len']):
                                extras_indptr[storage_node['extra'] + p] = extras_ptr_i
                                extras_val[extras_ptr_i: extras_ptr_i + node_val_len] = node_extras
                                extras_ptr_i += node_val_len

                        base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)
                        if base_children_len > 1:
                            if compute_node['extra'] != null_index:
                                loss_ptr_i, extras_ptr_i = first_time_layer_extra(compute_node['profile_len'], base_children_len, temp_children_queue,
                                                                                  nodes_array,
                                                                                  sidx_indptr, sidx_indexes,
                                                                                  loss_indptr, loss_val, loss_ptr_i,
                                                                                  extras_indptr, extras_val, extras_ptr_i
                                                                                  )
                            else:
                                loss_ptr_i = first_time_layer(compute_node['profile_len'], base_children_len, temp_children_queue,
                                                              nodes_array,
                                                              sidx_indptr, sidx_indexes,
                                                              loss_indptr, loss_val, loss_ptr_i
                                                              )

                    if children[storage_node['children']] and compute_node['extra'] != null_index:
                        # child is not base child so back allocation
                        # we need to keep track of extra before profile
                        for p in range(compute_node['profile_len']):
                            node_profile = node_profiles_array[compute_node['profiles'] + p]
                            if node_profile['i_start'] < node_profile['i_end']:
                                node_extras = extras_val[extras_indptr[storage_node['extra'] + p]:
                                                         extras_indptr[storage_node['extra'] + p] + node_val_len]

                                for i in range(node_val_len):
                                    temp_node_extras[p, node_sidx[i]] = node_extras[i]

            else:  # no children first level
                # we create space for extras and copy loss from layer 1 to other layer if they exist
                storage_node = compute_node
                node_sidx = sidx_val[sidx_indptr[sidx_indexes[storage_node['node_id']]]:
                                     sidx_indptr[sidx_indexes[storage_node['node_id']] + 1]]
                node_val_len = node_sidx.shape[0]
                node_loss = loss_val[loss_indptr[storage_node['loss']]:
                                     loss_indptr[storage_node['loss']] + node_val_len]

                if compute_node['extra'] != null_index:  # for layer 1 (p=0)
                    extras_indptr[storage_node['extra']] = extras_ptr_i
                    node_extras = extras_val[extras_ptr_i: extras_ptr_i + node_val_len]
                    node_extras.fill(0)
                    temp_node_extras[0].fill(0)
                    extras_ptr_i += node_val_len

                for p in range(1, compute_node['profile_len']):  # if base level already has layers
                    loss_indptr[storage_node['loss'] + p] = loss_ptr_i
                    loss_val[loss_ptr_i: loss_ptr_i + node_val_len] = node_loss
                    loss_ptr_i += node_val_len

                    if compute_node['extra'] != null_index:
                        extras_indptr[storage_node['extra'] + p] = extras_ptr_i
                        extras_val[extras_ptr_i: extras_ptr_i + node_val_len].fill(0)
                        extras_ptr_i += node_val_len
                        temp_node_extras[p].fill(0)

                for layer in range(compute_node['profile_len'], compute_node['layer_len']):
                    # fill up all layer if necessary
                    loss_indptr[storage_node['loss'] + layer] = loss_indptr[storage_node['loss']]
                    if compute_node['extra'] != null_index:
                        extras_indptr[storage_node['extra'] + layer] = extras_indptr[storage_node['extra']]

                if is_net_loss_or_allocation_rule_a1:
                    loss_indptr[storage_node['net_loss']] = loss_ptr_i
                    loss_val[loss_ptr_i: loss_ptr_i + node_val_len] = node_loss
                    loss_ptr_i += node_val_len

            base_children_len = 0

            # print('level', level, compute_node['agg_id'], loss_indptr[storage_node['loss']], loss_val[loss_indptr[storage_node['loss']]],
            #       node_profiles_array[compute_node['profiles']], compute_node['extra'] != null_index)

            # we apply the policy term on each layer
            for p in range(compute_node['profile_len']):
                node_profile = node_profiles_array[compute_node['profiles'] + p]
                if node_profile['i_start'] < node_profile['i_end']:
                    loss_in = loss_val[loss_indptr[storage_node['loss'] + p]:
                                       loss_indptr[storage_node['loss'] + p] + node_val_len]
                    loss_out = temp_node_loss_sparse[:node_val_len]

                    if compute_node['extra'] != null_index:
                        extra = extras_val[extras_indptr[storage_node['extra'] + p]:
                                           extras_indptr[storage_node['extra'] + p] + node_val_len]

                        for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                            calc_extra(fm_profile[profile_index],
                                       loss_out,
                                       loss_in,
                                       extra[:, DEDUCTIBLE],
                                       extra[:, OVERLIMIT],
                                       extra[:, UNDERLIMIT],
                                       stepped)
                            # print(compute_node['level_id'], 'fm_profile', fm_profile[profile_index])
                            # print(level, compute_node['agg_id'], base_children_len, p, fm_profile[profile_index]['calcrule_id'],
                            #       loss_indptr[storage_node['loss'] + p], loss_in, '=>', loss_out)
                            # print(temp_node_extras[p, node_sidx[0], DEDUCTIBLE], '=>', extra[0, DEDUCTIBLE], extras_indptr[storage_node['extra'] + p])
                            # print(temp_node_extras[p, node_sidx[0], OVERLIMIT], '=>', extra[0, OVERLIMIT])
                            # print(temp_node_extras[p, node_sidx[0], UNDERLIMIT], '=>', extra[0, UNDERLIMIT])
                        if not base_children_len:
                            base_children_len = get_base_children(storage_node, children, nodes_array,
                                                                  temp_children_queue)
                            if is_allocation_rule_a2:
                                ba_children_len = base_children_len
                            else:
                                ba_children_len = 1

                        back_alloc_extra_a2(ba_children_len, temp_children_queue, nodes_array, p,
                                            node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                                            loss_in, loss_out, temp_node_loss, loss_indptr, loss_val,
                                            extra, temp_node_extras, extras_indptr, extras_val)
                    else:
                        for profile_index in range(node_profile['i_start'], node_profile['i_end']):
                            calc(fm_profile[profile_index],
                                 loss_out,
                                 loss_in,
                                 stepped)
                            # print(compute_node['level_id'], 'fm_profile', fm_profile[profile_index])
                            # print(level, compute_node['agg_id'], base_children_len, p, fm_profile[profile_index]['calcrule_id'],
                            #       loss_indptr[storage_node['loss'] + p], loss_in, '=>', loss_out)
                        if not base_children_len:
                            base_children_len = get_base_children(storage_node, children, nodes_array,
                                                                  temp_children_queue)
                            if is_allocation_rule_a2:
                                ba_children_len = base_children_len
                            else:
                                ba_children_len = 1

                        back_alloc_a2(ba_children_len, temp_children_queue, nodes_array, p,
                                      node_val_len, node_sidx, sidx_indptr, sidx_indexes, sidx_val,
                                      loss_in, loss_out, temp_node_loss, loss_indptr, loss_val)

            if level != compute_info['max_level']:
                if compute_node['parent_len']:
                    parent_id = node_parents_array[compute_node['parent']]
                    next_compute_i = set_parent_next_compute(parent_id, storage_node['node_id'],
                                                             nodes_array, children, next_compute_i, computes)
                else:
                    if not base_children_len:
                        base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)

                    for base_child_i in range(base_children_len):
                        child = nodes_array[temp_children_queue[base_child_i]]
                        parent_id = node_parents_array[child['parent'] + item_parent_i[child['node_id']]]
                        item_parent_i[child['node_id']] += 1
                        next_compute_i = set_parent_next_compute(parent_id, child['node_id'],
                                                                 nodes_array, children, next_compute_i, computes)
            elif is_allocation_rule_a1:
                if not base_children_len:
                    base_children_len = get_base_children(storage_node, children, nodes_array, temp_children_queue)
                if base_children_len > 1:
                    temp_node_loss.fill(0)
                    for base_child_i in range(base_children_len):
                        child = nodes_array[temp_children_queue[base_child_i]]

                        child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
                        child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
                        child_val_len = child_sidx_end - child_sidx_start

                        child_sidx = sidx_val[child_sidx_start: child_sidx_end]
                        child_loss = loss_val[loss_indptr[child['net_loss']]: loss_indptr[child['net_loss']] + child_val_len]

                        for i in range(child_val_len):
                            temp_node_loss[:, child_sidx[i]] += child_loss[i]

                    for p in range(compute_node['profile_len']):
                        node_loss = loss_val[loss_indptr[storage_node['loss'] + p]:
                                             loss_indptr[storage_node['loss'] + p] + node_val_len]
                        for i in range(node_val_len):
                            if node_loss[i]:
                                temp_node_loss[p, node_sidx[i]] = node_loss[i] / temp_node_loss[p, node_sidx[i]]
                            else:
                                temp_node_loss[p, node_sidx[i]] = 0

                        for base_child_i in range(base_children_len):
                            child = nodes_array[temp_children_queue[base_child_i]]

                            child_sidx_start = sidx_indptr[sidx_indexes[child['node_id']]]
                            child_sidx_end = sidx_indptr[sidx_indexes[child['node_id']] + 1]
                            child_val_len = child_sidx_end - child_sidx_start

                            child_sidx = sidx_val[child_sidx_start: child_sidx_end]
                            child_loss = loss_val[loss_indptr[child['loss'] + p]: loss_indptr[child['loss'] + p] + child_val_len]
                            child_net = loss_val[loss_indptr[child['net_loss'] + p]: loss_indptr[child['net_loss'] + p] + child_val_len]

                            for i in range(child_val_len):
                                child_loss[i] = child_net[i] * temp_node_loss[p, child_sidx[i]]
            elif is_allocation_rule_a0:
                if compute_node['node_id'] != storage_node['node_id']:
                    sidx_indexes[compute_node['node_id']] = sidx_indexes[storage_node['node_id']]
                    for p in range(compute_node['profile_len']):
                        loss_indptr[compute_node['loss'] + p] = loss_indptr[storage_node['loss'] + p]

        compute_i += 1

    if net_loss:
        # we go through the last level node and replace the gross loss by net loss, then reset compute_i to its value
        net_compute_i = 0
        while computes[net_compute_i]:
            node_i, net_compute_i = computes[net_compute_i], net_compute_i + 1
            node = nodes_array[node_i]
            # net loss layer i is initial loss - sum of all layer up to i
            node_val_len = sidx_indptr[sidx_indexes[node['node_id']] + 1] - sidx_indptr[sidx_indexes[node['node_id']]]
            node_ba_val_prev = loss_val[loss_indptr[node['net_loss']]: loss_indptr[node['net_loss']] + node_val_len]
            for layer in range(node['layer_len']):
                node_ba_val_cur = loss_val[loss_indptr[node['loss'] + layer]: loss_indptr[node['loss'] + layer] + node_val_len]
                # print(node['agg_id'], layer, loss_indptr[node['loss'] + layer], node_ba_val_prev, node_ba_val_cur)
                node_ba_val_cur[:] = np.maximum(node_ba_val_prev - node_ba_val_cur, 0)
                node_ba_val_prev = node_ba_val_cur

    item_parent_i.fill(1)
    # print(compute_info['max_level'], next_compute_i, compute_i, next_compute_i-compute_i, computes[compute_i:compute_i + 2], computes[next_compute_i - 1: next_compute_i + 1])
    if compute_info['allocation_rule'] == 0:
        return level_start_compute_i
    else:
        return compute_i


def init_variable(compute_info, max_sidx_val, temp_dir, low_memory):
    """
    extras, loss contains the same index as sidx
    therefore we can use only sidx_indexes to tract the length of each node values
    """
    max_sidx_count = max_sidx_val + EXTRA_VALUES
    len_array = max_sidx_val + 6

    if low_memory:
        sidx_val = np.memmap(os.path.join(temp_dir, "sidx_val.bin"), mode='w+',
                             shape=(compute_info['node_len'] * max_sidx_count), dtype=np_oasis_int)
        loss_val = np.memmap(os.path.join(temp_dir, "loss_val.bin"), mode='w+',
                             shape=(compute_info['loss_len'] * max_sidx_count), dtype=np_oasis_float)
        extras_val = np.memmap(os.path.join(temp_dir, "extras_val.bin"), mode='w+',
                               shape=(compute_info['extra_len'] * max_sidx_count, 3), dtype=np_oasis_float)
    else:
        sidx_val = np.zeros((compute_info['node_len'] * max_sidx_count), dtype=np_oasis_int)
        loss_val = np.zeros((compute_info['loss_len'] * max_sidx_count), dtype=np_oasis_float)
        extras_val = np.zeros((compute_info['extra_len'] * max_sidx_count, 3), dtype=np_oasis_float)

    sidx_indptr = np.zeros(compute_info['node_len'] + 1, dtype=np.int64)
    loss_indptr = np.zeros(compute_info['loss_len'] + 1, dtype=np.int64)
    extras_indptr = np.zeros(compute_info['extra_len'] + 1, dtype=np.int64)

    sidx_indexes = np.empty(compute_info['node_len'], dtype=np_oasis_int)
    children = np.zeros(compute_info['children_len'], dtype=np.uint32)
    computes = np.zeros(compute_info['compute_len'], dtype=np.uint32)

    pass_through = np.zeros(compute_info['items_len'] + 1, dtype=np_oasis_float)
    item_parent_i = np.ones(compute_info['items_len'] + 1, dtype=np.uint32)

    return (max_sidx_val, max_sidx_count, len_array, sidx_indexes, sidx_indptr, sidx_val, loss_indptr, loss_val,
            pass_through, extras_indptr, extras_val, children, computes, item_parent_i)


@njit(cache=True)
def reset_variable(children, compute_i, computes):
    computes[:compute_i].fill(0)
    children.fill(0)



import datetime
import os.path

from .path import get_custom_module
from .exceptions import OasisException

from ods_tools.oed import DictBasedCurrencyRates, get_currency_col, convert_currency
try:
    from forex_python.converter import CurrencyRates as BaseFxCurrencyRates
    from forex_python.converter import RatesNotAvailableError

    class FxCurrencyRates(BaseFxCurrencyRates):
        def __init__(self, *args, date_obj=None, **kwargs):
            super().__init__(*args, **kwargs)
            self.date_obj = date_obj

        def get_rate(self, base_cur, dest_cur, date_obj=None):
            if date_obj is None:
                date_obj = self.date_obj
            try:
                return super().get_rate(base_cur, dest_cur, date_obj)
            except RatesNotAvailableError as e:
                raise OasisException("Issue retrieving rate, most probably the forex api is down") from e

except ImportError:
    FxCurrencyRates = None

import logging
logger = logging.getLogger(__name__)


def create_currency_rates(currency_conversion):
    def get_path(name):
        if os.path.isabs(currency_conversion[name]):
            return currency_conversion[name]
        else:
            return os.path.join(currency_conversion['root_dir'], currency_conversion[name])

    if currency_conversion.get('currency_conversion_type') == 'DictBasedCurrencyRates':
        if currency_conversion.get("source_type") == 'csv':
            return DictBasedCurrencyRates.from_csv(get_path('file_path'), **currency_conversion.get("read_parameters", {}))
        elif currency_conversion.get("source_type") == 'parquet':
            return DictBasedCurrencyRates.from_parquet(get_path('file_path'), **currency_conversion.get("read_parameters", {}))
        elif currency_conversion.get("source_type", '').lower() == 'dict':
            return DictBasedCurrencyRates(currency_conversion['currency_rates'])
        else:
            raise OasisException(f"Unsuported currency_conversion source type : {currency_conversion.get('source_type')}")
    elif currency_conversion.get('currency_conversion_type') == 'FxCurrencyRates':
        if FxCurrencyRates is None:
            raise OasisException("You must install package forex-python to use builtin_currency_conversion_type FxCurrencyRates")

        _datetime = currency_conversion.get('datetime')
        if _datetime is not None:
            _datetime = datetime.datetime.fromisoformat(_datetime)
        return FxCurrencyRates(date_obj=_datetime, **currency_conversion.get("fx_currency_rates_parameters", {}))
    elif currency_conversion.get('currency_conversion_type') == 'custom':
        _module = get_custom_module(get_path('module_path'), 'Currency Rate lookup module path')

        try:
            _class = getattr(_module, currency_conversion['class_name'])
        except AttributeError as e:
            raise OasisException(f"class {currency_conversion['class_name']} "
                                 f"is not defined in module {get_path('module_path')}") from e.__cause__

        return _class(**currency_conversion.get("custom_parameters", {}))

    else:
        raise OasisException(f"unsupported currency_conversion_type {currency_conversion.get('currency_conversion_type')}")


def manage_multiple_currency(oed_path, oed_df, reporting_currency, currency_rate, ods_fields):
    currency_col = get_currency_col(oed_df)
    currencies = oed_df[currency_col].unique()

    if not reporting_currency:
        if len(currencies) > 1:
            logger.warning(f"file {oed_path} contains multiple currencies, but no reporting_currency has been specified")

    elif currency_rate is None:
        if len(currencies) > 1 or currencies[0] != reporting_currency:
            raise OasisException(f"currency_conversion_json is necessary to perform conversion of {oed_path} to {reporting_currency}")

    else:
        logs = []
        for cur in currencies:
            if cur != reporting_currency:
                logs.append(f"{cur} => {reporting_currency}: rate of exchange {currency_rate.get_rate(cur, reporting_currency)}")
        if logs:
            logger.info(f"currency conversion for file {oed_path}:\n\t" + "\n\t".join(logs))
            convert_currency(oed_df, reporting_currency, currency_rate, ods_fields)
            return True
    return False



"""
This file defines the functions that maps the supported perils with the correlation settings. This data is usually
obtained from the model_settings.
"""
import pandas as pd
from typing import Optional


def map_data(data: Optional[dict]) -> Optional[pd.DataFrame]:
    """
    Maps data from the model settings to to have Peril ID, peril_correlation_group, and correlation_value.

    Args:
        data: (dict) the data loaded from the model settings

    Returns: (pd.DataFrame) the mapped data
    """
    if data is not None:
        supported_perils = data.get("lookup_settings", {}).get("supported_perils", [])
        correlation_settings = data.get("correlation_settings", [])

        for supported_peril in supported_perils:
            supported_peril["peril_correlation_group"] = supported_peril.get("peril_correlation_group", 0)

        supported_perils_df = pd.DataFrame(supported_perils)
        correlation_settings_df = pd.DataFrame(correlation_settings)

        # merge allows duplicates of the "peril_correlation_group" in the supported perils
        # merge does not allow duplicates of the "peril_correlation_group" in the correlation settings
        if len(supported_perils_df) > 0 and len(correlation_settings_df) > 0:
            mapped_data = pd.merge(supported_perils_df, correlation_settings_df, on="peril_correlation_group")
            return mapped_data


def get_correlation_input_items(gul_inputs_df: pd.DataFrame, correlation_map_df: pd.DataFrame) -> pd.DataFrame:
    """
    Gets the correlation values with the peril ID from the model_settings.

    Args:
        correlation_map_df: (pd.DataFrame) data from the model settings to to have Peril ID, peril_correlation_group,
                                           and correlation_value
        gul_inputs_df: (pd.DataFrame) the data of the gul inputs to be mapped

    Returns: (pd.DataFrame) the mapped data of correlations
    """
    gul_inputs_df = gul_inputs_df.merge(correlation_map_df, left_on='peril_id', right_on='id').reset_index()
    gul_inputs_df["correlation_value"] = gul_inputs_df["correlation_value"].astype(float)
    gul_inputs_df = gul_inputs_df.reindex(columns=list(gul_inputs_df))

    correlation_df = gul_inputs_df[["item_id", "peril_correlation_group", "correlation_value"]]
    return correlation_df.sort_values('item_id')
